import subprocess
from shlex import split

class ShellException(OSError):
    pass

def run(command: str) -> list[str]:
    result = subprocess.run(split(command), stdout=subprocess.PIPE)
    stdout = format(result.stdout)

    if not ok(result):
        raise ShellException(stdout)

    return stdout

def format(stream: bytes) -> list[str]:
    return stream.decode('utf-8').splitlines()

def ok(result: subprocess.CompletedProcess) -> bool:
    return result.returncode == 0


"""Git operations"""

from publicator import shell

def current_branch() -> str:
    return shell.run("git symbolic-ref --short HEAD").pop()

def release_branches() -> tuple[str, str]:
    return ("main", "master")

def status() -> list[str]:
    return shell.run("git status --porcelain")

def is_working_directory_clean() -> bool:
    return len(status()) == 0

def stash() -> list[str]:
    return shell.run("git stash -u")

def pull() -> list[str]:
    return shell.run("git pull --rebase")

def pop() -> list[str]:
    return shell.run("git stash pop")

def add() -> list[str]:
    return shell.run("git add pyproject.toml")

def commit(message: str) -> list[str]:
    return shell.run(f'git commit -m "{message.strip()}"')

def create_tag(version: str, message: str) -> list[str]:
    return shell.run(f'git tag -a {version.strip()} -m "{message.strip()}"')

def push() -> list[str]:
    return shell.run("git push --follow-tags")



from __future__ import annotations
from dataclasses import dataclass
from unittest.mock import patch

@dataclass(frozen=True)
class Semver():
    major: int = 0
    minor: int = 1
    patch: int = 0

    @classmethod
    def from_string(self, version: str) -> Semver:
        major, minor, patch = map(int, version.split('.'))
        return Semver(major, minor, patch)

    def as_tuple(self) -> tuple[int]:
        return (self.major, self.minor, self.patch)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Semver):
            return self.as_tuple() == other.as_tuple()

        return False

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}.{self.patch}"

    def __repr__(self) -> str:
        return f"Version ({self})"



from typing import Any
import toml
from pathlib import Path
from mergedeep import merge

from publicator.semver import Semver

def get_version() -> Semver:
    pyproject = Path.cwd() / "pyproject.toml"
    contents: dict[str, Any] = toml.loads(pyproject.read_text())
    version = contents['tool']['poetry']['version']

    return Semver.from_string(version)

def bump_version(version: str) -> Semver:
    pyproject = Path.cwd() / "pyproject.toml"

    contents: dict[str, Any] = toml.loads(pyproject.read_text())
    bump = {
        "tool": {
            "poetry": {
                "version": version
            }
        }
    }

    new_contents = merge({}, contents, bump)

    pyproject.write_text(toml.dumps(new_contents))

    return Semver.from_string(version)



import os
from typing import Optional
import typer

from publicator import git, poetry, project
from publicator.semver import Semver

preview = os.environ.get("PUBLICATOR_PREVIEW")

app = typer.Typer(name="publicator")

@app.command()
def cli(
    version: str = typer.Argument(..., metavar="version", help="can be one of (patch | minor | major | 1.2.3)"),
    repository: Optional[str] = typer.Option(default=None, metavar="name", help="Custom repository for publishing (must be specified in pyproject.toml)"),
    any_branch: bool = typer.Option(default=False, help="Allow publishing from any branch"),
    skip_cleaning: bool = typer.Option(default=False, help="Skip repository clean up"),
    yolo: bool = typer.Option(default=False, help="Skip reinstall and test steps"),
    skip_tag: bool = typer.Option(default=False, help="Skip creating a new tag"),
    skip_publish: bool = typer.Option(default=False, help="Skip publishing the package to the registry"),
    skip_push: bool = typer.Option(default=False, help="Skip pushing commits and tags to Git"),
) -> None:
    if not any_branch:
        verify_branch()

    if not skip_cleaning:
        clean_up()

    if not yolo:
        install_dependencies()
        run_tests()

    semver = bump_version(version)
    commit_changes(semver)

    if not skip_tag:
        create_tag(version)

    build_package()

    if not skip_publish:
        publish_package(repository, skip_publish)

    if not skip_push:
        push()

    success(f"Published the new package version. Cheers!")

def info(message: str) -> None:
    typer.secho(f"📢 {message}", fg=typer.colors.BRIGHT_BLUE)

def success(message: str) -> None:
    typer.secho(f"🎉 {message}", fg=typer.colors.BRIGHT_GREEN, bold=True)

def fatal(message: str, exit_code: int = 1) -> None:
    typer.secho(f"❌ {message}", fg=typer.colors.BRIGHT_RED, bold=True)
    raise typer.Exit(code=exit_code)

def push():
    info("Pushing changes to Git")
    if not preview:
        git.push()

def publish_package(repository: str, skip_publish: bool) -> None:
    info("Publishing the package to repository")
    if not preview:
        poetry.publish(repository, dry_run=skip_publish)

def build_package() -> None:
    info("Building the package")
    if not preview:
        poetry.build()

def create_tag(version: Semver) -> None:
    info(f"Creating a new tag {version} from HEAD")
    if not preview:
        git.create_tag(version, message=f"Version {version}")

def commit_changes(semver: Semver) -> None:
    info("Committing changes")
    if not preview:
        poetry.ok()
        git.add()
        git.commit(f"release: {semver}")

def bump_version(version: str) -> Semver:
    current_version = project.get_version()
    info(f"Bumping current version from {current_version} to {version}")

    if preview:
        return current_version

    return project.bump_version(version)

def run_tests():
    info("Running tests")
    if not preview:
        poetry.run_tests()

def install_dependencies():
    info("Reinstalling dependencies")
    if not preview:
        poetry.install()

def clean_up() -> None:
    if git.is_working_directory_clean():
        return

    if not preview:
        info("Resetting working directory to a clean state")
        git.stash()
        git.pull()
        git.pop()

def verify_branch() -> None:
    current_branch = git.current_branch()
    release_branches = git.release_branches()

    if not current_branch in release_branches:
        fatal(f"Current checked out branch {current_branch} is not a release branch {release_branches}")



from typing import Optional
from publicator import shell


def install() -> list[str]:
    return shell.run("poetry install --remove-untracked")

def run_tests(command: str = "pytest") -> list[str]:
    return shell.run(f"poetry run {command.strip()}")

def ok() -> bool:
    result = shell.run("poetry check").pop()
    return "All set" in result

def build() -> list[str]:
    return shell.run("poetry build")

def publish(repository: Optional[str] = None, dry_run: bool = False) -> list[str]:
    command = "poetry publish"

    if repository:
        command += f" --repository {repository}"
    if dry_run:
        command += f" --dry-run"

    return shell.run(command)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['publicator']

package_data = \
{'': ['*']}

install_requires = \
['mergedeep>=1.3.4,<2.0.0', 'toml>=0.10.2,<0.11.0', 'typer[all]>=0.4.0,<0.5.0']

entry_points = \
{'console_scripts': ['publicator = publicator.cli:app']}

setup_kwargs = {
    'name': 'publicator',
    'version': '0.1.3',
    'description': '',
    'long_description': "# Publicator\n\n> A better `poetry publish`\n\n## Features\n\n### Supported\n\n- Ensures you are publishing from your release branch (main and master by default)\n- Ensures the working directory is clean and that there are no changes not pulled\n- Reinstalls dependencies to ensure your project works with the latest dependency tree\n- Ensures your Python version is supported by the project and its dependencies\n- Runs the tests\n- Bumps the version in pyproject.toml and creates a git tag\n- Publishes the new version to pypi.org\n- Pushes commits and tags (newly & previously created) to your Git server\n\n### Upcoming\n\n- Interactive UI\n- Opens a prefilled GitHub Releases draft after publish\n- See exactly what will be executed with preview mode, without pushing or publishing anything remotely\n\n## Prerequisites\n\n- Python 3.9 or later\n- Poetry 1.1 or later\n- Git 2.11 or later\n\n## Install\n\nInstall with pipx.\n\n```sh\npipx install publicator\n```\n\n## Usage\n\n```plain\n$ publicator --help\n\nUsage:\n\n$ publicator <version>\n\nVersion can be one of:\n    patch | minor | major | 1.2.3\n\nOptions\n    --any-branch            Allow publishing from any branch\n    --branch                Name of the release branch (default: main | master)\n    --no-tests              Skips tests\n    --no-publish            Skips publishing\n    --preview               Show tasks without actually executing them\n    --no-release-draft      Skips opening a GitHub release draft\n    --release-draft-only    Only opens a GitHub release draft\n    --test-script           Name of shell command to run tests before publishing (default: `poetry run pytest`)\n    --message               Version bump commit message. `%s` will be replaced with version. (default: '%s')\n```\n\n## Configuration\n\nTo be added later.\n",
    'author': 'Niko Heikkilä',
    'author_email': 'niko.heikkila@futurice.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
